
# Distribution.py
# class definition of the Distribution class
# and supporting functions

from operator import itemgetter
import numpy as np
import itertools
import cPickle as pickle

def LPWMI(D):
    return float(D[0,0])*np.log(float(D[0,0] / (D[0,:].sum() * D[:,0].sum())))

def inc_exc(observables, moments, loc):
    zeros = [x for x, y in enumerate(loc) if y == 0]
    ones = [x for x, y in enumerate(loc) if y == 1]
    P = moments[frozenset(zeros)]
    for m in xrange(len(ones)):
        sign = (-1)**(m+1)
        p = 0
        for subset in itertools.combinations(ones, m+1):
            if len(list(subset)+zeros) >= 2:
                p += moments[frozenset(itemgetter(*(zeros+list(subset)))(observables))]
            elif len(list(subset)+zeros) == 1:
                p += moments[frozenset(list([itemgetter(*(zeros+list(subset)))(observables)]))]
            elif len(list(subset)+zeros) == 0:
                p += moments[frozenset()]

        P += sign*p
    return P

def argsort(seq):
    # http://stackoverflow.com/questions/3071415/efficient-method-to-calculate-the-rank-vector-of-a-list-in-python
    return sorted(range(len(seq)), key = seq.__getitem__)

def toMoments(T):
    K = len(T.shape)
    M = {}
    for k in xrange(K+1):
        for obs in itertools.combinations(range(K), k):
            C = list(set(range(K)) - set(obs))
            S = np.apply_over_axes(np.sum, T, C)
            M[frozenset(obs)] = S.flatten()[0]
    return M
    
def toTensor(M):
    K = max([len(z) for z in M.keys()])
    T = np.zeros((2,)*K)
    for loc in itertools.product([0,1], repeat=K):
        T[loc] = inc_exc(range(K), M, loc)
    return T


def readData(network_name, schedule, samples, seed):
    if samples == 'inf':
        return read_params_from_dir(network_name)
    else:
        
        infile = file(network_name + "/samples/sufficientStatistics_n"+str(samples)+"_s"+str(seed)+"_"+schedule)
        quartet_ids = file(network_name + "/samples/id_n"+str(samples)+"_s"+str(seed)+"_"+schedule).readlines()

        print "reading data from", infile.name
        numChildren = int(infile.readline())
        data = {}
        i = 0
        for l in infile:
            l = l.strip().split()
            dat = [int(x) for x in l]
            iden = tuple([int(x) for x in quartet_ids[i].strip().split()])
            dat =  np.array(dat).reshape((2,2,2,2))
            dat =  dat / float(dat.sum())
            data[iden] = dat
            i+=1
            if i % 1000 == 0:
                print i
        print 'done'
        return numChildren, data
    
def read_params_from_dir(dir_name):
    
    noise = np.array([float(x) for x in file(dir_name+"/true_params/noise.txt").read().rstrip().split('\n')][1:])
    priors = np.array([float(x) for x in file(dir_name+"/true_params/priors.txt").read().rstrip().split('\n')][1:])
    
    
    edges = {}
    f = file(dir_name+"/true_params/weights.txt")
    i = 0
    for l in f:
        l = [float(x) for x in l.split()]
        for j in xrange(len(l)):
            if l[j] < 1 - 10**-9:
                edges[(i, j)] = float(l[j])
        i += 1

    for i in xrange(priors.size):
        edges[(i, None)] = priors[i]
    for j in xrange(noise.size):
        edges[(None, j)] = noise[j]

    for e in edges:
        print e, edges[e]
        assert(edges[e] >= 0 and edges[e] <= 1)

    print "read priors", priors
    print "read noise", noise
    print "read edges", edges
    return edges

def samples_from(D, numSamples):
    samples = np.zeros(D.shape)
    for _ in xrange(100):
        P = D.flatten()
        samples += (np.random.multinomial(numSamples, P)).reshape(D.shape) / float(numSamples)
    
    samples = samples / 100
    return samples

class Distribution:

    def __init__(self, network_name, schedule, samples, seed, prune_pairs, checkpoint_prefix, npairs=1500, blacklist = None):
        self.seed = seed
        if blacklist:
            self.blacklist = set(blacklist)
        else:
            self.blacklist = set()

        self.npairs = npairs
        self.numSamples = samples
        data = readData(network_name, schedule, samples, self.seed)
            
        if samples == 'inf':
            #print "infinite samples"
            self.edges = data
            self.parents_of = {}
            for p,c in self.edges:
                if (not p==None) and (not c == None):
                    if not c in self.parents_of:
                        self.parents_of[c] = set()
                    self.parents_of[c].add(p)

            self.numChildren = len(self.parents_of)
        else:
            self.numChildren, self.sampled_dist = data
        
        self.siblings_of = self.initialize_true_siblings(network_name)
        print "initialized distribution"

    def true_distribution(self, observables):
        M = {}
        K = len(observables)
        for k in xrange(K+1):
            for indices in itertools.combinations(range(K), k):
                obs = [observables[i] for i in indices]
                M[frozenset(indices)] = self.getPerfectMoment(obs)
        D = toTensor(M)
        assert np.abs(D.sum() - 1) < 10**-9
        assert (D.flatten() > 0).all()
        return D

    def initialize_true_siblings(self, network_name):
        siblings_of = {}
    
        pairings = []
        print 'num children', self.numChildren
        for (a,b) in itertools.combinations(xrange(self.numChildren), 2):
            pairings.append((LPWMI(self.empirical_distribution((a,b))), (a,b)))

        pairings.sort(reverse=True)
        print "there are", len(pairings), "pairs"
        for  _, (a,b) in pairings:
            if not a in siblings_of:
                siblings_of[a] = set()
            if not b in siblings_of:
                siblings_of[b] = set()

        #pairings = pairings[:self.npairs] 
        print "now there are", len(pairings), "pairs"

        for  _, (a,b) in pairings:
            siblings_of[a].add(b)
            siblings_of[b].add(a)

        return siblings_of
        

    def listSets(self, s):
        if self.numSamples == 'inf':
            #print "listing sets", self.numChildren
            return [z for z in itertools.combinations(xrange(self.numChildren), s)]
        else:
            print 'sampled_dist.keys()', len(self.sampled_dist.keys())
            sets = [z[:s] for z in filter(lambda x: 4 - x.count(-1) == s, self.sampled_dist.keys())]
            print 'before pruning', len(sets)
            if s ==4:
                sets = filter(self.do_pruning, sets)
                for s in sets:
                    assert len(set(s) & self.blacklist) == 0
            print 'after pruning', len(sets)

            return sets

    def do_pruning(self, s):
        if len(self.blacklist & set(s)): #get rid of blacklisted variables
            return False

        if len(s) == 1:
            return True

        #print s,
        for a,b in itertools.combinations(s,2):
            if not (a in self.siblings_of[b]):
                #print 'rejected', a, 'not a sibling of', b
                return False
        else:
            #print 'accepted'
            return True

    def empirical_distribution(self, obs):
        '''empirical_distribution --
        returns the joint distribution over the variables
        specified in the obs argument.'''

        if self.numSamples == 'inf':
            return self.true_distribution(obs)
        else:
            l = len(obs)

            #reorder the arguments
            order = argsort(obs)
            obs = tuple(sorted(obs))

            #pad with -1
            while(len(obs) < 4):
                obs = obs+(-1,)

            #look up the distribution from the dictionary
            D = self.sampled_dist[obs] 

            #resize the array to be the requested size
            while(len(D.shape) > l):
                D = D.sum(len(D.shape)-1)
            
            #reshuffle the array to be in the requested order
            D = np.transpose(D, order)

            assert np.abs(D.sum() - 1) < 10**-9
            return D

    def getPerfectMoment(self, obs):
        '''getPerfectMoment -- function to find
        what the moments would be with infinite data'''
        parents = set()
        for o in obs:
            parents |= self.parents_of[o]

        K = len(parents) 
        N = len(obs)

        parents = list(parents)

        noise = np.zeros(N)
        failures = np.zeros((K, N))
        priors = np.zeros(K)

        for k in xrange(K):
            priors[k] = self.edges[(parents[k], None)]
            for i in xrange(N):
                if (parents[k], obs[i]) in self.edges:
                    failures[k,i] = self.edges[(parents[k], obs[i])]
                else:
                    failures[k,i] = 1

        for i in xrange(N):
            noise[i] = self.edges[(None,obs[i])]

        M = 1
        for i in xrange(N):
            M *= noise[i]

        for k in xrange(K):
            f = 1
            for i in xrange(N):
                f *= failures[k,i]
            M*= (1-priors[k] + priors[k]*f)
        return M
