
# Structure_learner.py
# Class definition and supporting functions for the 
# Structure_learner class

import time
from multiprocessing import Pool
import numpy as np
from Decomp import learnParams_decompose as do_learnParams_decompose
from Distribution import toTensor, toMoments
import itertools

def decompose((D, S)):
    if D == None:
        return (None, None, None)
    return do_learnParams_decompose(D, include_mixture=0)

def global_testWeaklyCoupled_eigen(D, children):
    if D == None:
        return float('inf'), 'none'
    if np.min(D) < 0:
        return float('inf'), 'neg', np.min(D)
        
    Q = D.reshape((4,4))

    eigs = []
    
    _, e, _ = np.linalg.svd(Q)
    eigs.append(np.sort(e)[1])

    Q = np.transpose(D, [0,2,1,3]).reshape((4,4))

    _, e, _ = np.linalg.svd(Q)
    eigs.append(np.sort(e)[1])
    
    Q = np.transpose(D, [0,3,1,2]).reshape((4,4))

    _, e, _ = np.linalg.svd(Q)
    eigs.append(np.sort(e)[1])
        
    return max(eigs)

def global_testCoupled(D, (a,b,c)):
    conditional_D = D[:,:,0] / (D[:,:,0]).sum()
    unconditional_D = D.sum(2)
    conditional_PWMI = PWMI(conditional_D)
    unconditional_PWMI = PWMI(unconditional_D)
    coupled_score = unconditional_PWMI - conditional_PWMI

    return coupled_score

def parallelEigenTest(args):

    (D, adjusted), children, round = args
    a,b,c,d = children

    if round > 0 and not adjusted:
        return(a,b,c,d), (float('-inf'),float('-inf'), float('-inf'), float('-inf'), float('inf'), "not adjusted"), D

    if D == None:
        return (a,b,c,d), (float('-inf'),float('-inf'), float('-inf'), float('-inf'), float('inf'), "D not available"), D
    
    c1 = global_testCoupled(D.sum(0), (b,c,d))
    c2 =  global_testCoupled(D.sum(1), (a,c,d))
    c3 = global_testCoupled(D.sum(2), (a,b,d))
    c4 = global_testCoupled(D.sum(3), (a,b,c))

    score = global_testWeaklyCoupled_eigen(D, (a,b,c,d))

    return (a,b,c,d), (c1, c2, c3, c4, score), D


def median(l):
    return sorted(l)[len(l)/2]

def select_best(l):
    scores = [np.std(z[1:]) for z in l]
    return l[np.argmin(scores)][0]

def select_median(l):
    vals = [z[0] for z in l]
    if np.isnan(vals).any():
        return float('NaN'), float('NaN')
    if np.isinf(vals).any():
        return float('NaN'), float('NaN')
    return np.median(vals), np.std(vals)
    
def solve_quad(a,b,c):
    r = (-b - np.lib.scimath.sqrt(b**2 - 4*a*c))/(2*a)
    q = (-b + np.lib.scimath.sqrt(b**2 - 4*a*c))/(2*a)
    
    if b**2 - 4*a*c < 0:
        #print "COMPLEX"
        #print q, r
        q = -b/float(2*a)
        r = q
        #print "using min", q, "instead of root"

    if(q > 0.5):
        pass
        #print "WARNING: root is > 0.5", q

    return q, r

def sampled(D, n):
    assert np.abs(D.sum() - 1) < 10**-9
    s = np.random.multinomial(n, D.flatten()/float(D.sum())).reshape(D.shape)
    s = s / float(s.sum())
    assert np.abs(s.sum() - 1) < 10**-9
    return s
    

#pointwise mutual information
def PWMI(D):
    return float(D[0,0] / (D[0,:].sum() * D[:,0].sum()))

def LPWMI(D):
    return float(D[0,0])*np.log(float(D[0,0] / (D[0,:].sum() * D[:,0].sum())))


class Structure_Learner:
    def __init__(self, children, dist, network_name, checkpoint_prefix, results_dir):


        self.children = children
        self.currentParent = 0

        self.dist = dist
        self.numChildren = len(children)
        self.network_name = network_name
        
        self.edges = {}
        self.certainties = {}
        self.children_of = {}
        self.parents_of = {}
        self.parents_in_round = {}
        for c in children:
            self.parents_of[c] = set()
        self.singly_coupled = {}
        self.checkpoints = checkpoint_prefix
        self.results_dir = results_dir


    #########################
    #Checking Relationships #
    #########################

    def initialize_siblings(self):
        #print "checking siblings"
        self.siblings = {}
        for a in self.children:
            #print "siblings of", a
            for b in self.children[a+1:]:
                if self.testSiblings(a,b):
                    if not a in self.siblings:
                        self.siblings[a] = set()
                    if not b in self.siblings:
                        self.siblings[b] = set()

                    self.siblings[a].add(b)
                    self.siblings[b].add(a)
        
    def testSiblings(self, a,b):
        try:
            D = self.dist.empirical_distribution((a,b))
        except:
            return False
        else:
            return True 

    def couples(self, p, children):
        coupling_parents = set()
        for (a,b) in itertools.combinations(children,2):
            coupling_parents |= (self.parents_of[a] & self.parents_of[b])
        return (p in coupling_parents)

    ##########################
    #Adjusting Distributions #
    ##########################
    def adjust(self, D, children, subtract_off_from_these=None, verbose=False):
        if subtract_off_from_these == None:
            subtract_off_from_these = children

        needed = False
        parents = range(self.currentParent)
        for p in parents:
            if p in self.parents_in_round[self.current_round]:
                continue
            if self.couples(p, subtract_off_from_these):
                needed = True
                break

        if needed:
            M = toMoments(D) 
            for p in parents:
                if p in self.parents_in_round[self.current_round]:
                    continue

                if self.couples(p, subtract_off_from_these):
                    if verbose:
                        print "subtracting off", p, "from", children
                        print "M before:", M
                    M = self.remove(M, children, p) 
                    if verbose:
                        print "M after:", M

            if verbose:
                print "D before", D

            D = toTensor(M)

            if verbose:
                print "D after", D

        return D, needed

    def remove(self, M, children, pa):
        p = self.edges[(pa, None)]

        for m in M.keys():
            f = 1
            for i in m:
                if children[i] in self.children_of[pa]:
                    f *= self.edges[(pa, children[i])]
        
            M[m] = M[m] / float(1-p + p*f)

        return M

    def adjustedD(self, S):
        return self.adjust(self.dist.empirical_distribution(S), S)

    ################
    #Decomposition #
    ################

    def learnParams(self, D, children):
        children = tuple(sorted(children))
        priors = []
        failures = [[] for _ in xrange(3)]

        
        p, f, _ = decompose((D, children))

        priors.append(p)
        for i in xrange(3):
            failures[i].append(f[i])

        p, f, _ = decompose((D, children))
        priors.append(p)
        for i in xrange(3):
            failures[i].append(f[i])

        return priors, failures, _


    #################
    #Extending Step #
    #################



    def extend(self, pa, (a,b), z):
        f_a = self.edges[(pa, a)]
        f_b = self.edges[(pa, b)]
        p = self.edges[(pa, None)]

        try:
            D = self.dist.empirical_distribution((a,b,z))
            assert not (D == None)
        except:
            print "could not get distribution"
            return float('nan')

        try:
            D, _ = self.adjust(D, (a,b,z), (a,b))
            assert not (D == None)
        except:
            print "could not adjust distribution"
            return float('nan')

        conditional_D = D[:,:,0] / D[:,:,0].sum()

        R = PWMI(conditional_D)
        
        if (PWMI(D.sum(2)) - R < self.coupled_thresh):
            #print "did not pass coupled threshold",  PWMI(D.sum(2)) - R 
            return float('nan')

        else:
            #print "passed coupled threshold", PWMI(D.sum(2)) - R , "!"
            quad_A = -f_a*f_b*R + f_a*R + f_b*R - R
            quad_B = f_a*f_b - f_a*R - f_b*R + 2*R - 1
            quad_C = 1-R
            #p_nz is p conditioned on not-z
            p_nz, other = solve_quad(quad_A, quad_B, quad_C)
            f_z = (p_nz*(1-p))/((1-p_nz)*p)

        return f_z

    def getBestPair(self, pa, children, z, opt):
        '''getBestPair - decide which pair to use in extend step'''

        #option 1: global best pair
        if opt == 'coherence':
            v, (a,b,c,d) = zip(*sorted(zip([self.certainties[(pa, z)] for z in children], children), key=lambda x:x[0]))
        elif opt == 'marginal':
            D = self.dist.empirical_distribution(children)
            marginals = [D[0,:,:,:].sum(), D[:,0,:,:].sum(), D[:, :, 0, :].sum(), D[:, :, :, 0].sum()]
            v, (a,b,c,d) = zip(*sorted(zip(marginals, children), key=lambda x: np.abs(x[0] - 0.5)))
        return a,b
        
    def extendParent(self, pa, z, children):
        for (a,b) in itertools.combinations(children, 2):
            f_z = self.extend(pa, (a,b), z)
            if np.isnan(f_z):
                return False

        a,b = self.getBestPair(pa, children, z, 'coherence')

        f_z = self.extend(pa, (a,b), z)
        if np.isnan(f_z):
            return False

        val = np.clip(f_z, self.f_min, self.f_max)

        self.edges[(pa, z)] = val
        return True

    
    def learnNewParent(self, children, D, r):


        for (a,b) in itertools.combinations(children,2):
            if len(self.parents_of[a] & self.parents_of[b] & self.parents_in_round[r]):
                print 'covered by previous parents'
                return -1

        prior_est = []
        failure_est = [[] for _ in xrange(4)]

        for i in xrange(4):
            prior, failures, _ = self.learnParams(D.sum(i), sorted(tuple(set(children) - set((children[i], )))))
            prior_est.append(prior)
            for j in xrange(4):
                if j < i:
                    failure_est[j].append(failures[j])
                if j > i:
                    failure_est[j].append(failures[j-1])
                
        p = self.currentParent

        val,var = select_median(prior_est)

        if val < self.p_min or val > self.p_max:
            print 'out of range p', val
            return -1

        if np.isnan(val):
            print 'nan p'
            return -1
            
        for i in xrange(4):
            val, var = select_median(failure_est[i])
            if np.isnan(val):
                print 'nan f', val
                return -1
            if val < self.f_min or val > self.f_max:
                print 'out of range f', val
                return -1

        val,var = select_median(prior_est)
        self.edges[(p, None)] = val
        self.certainties[(p,None)] = var

        for i in xrange(4):
            
            val, var = select_median(failure_est[i])
            self.edges[(p,children[i])] = np.clip(val, 0, 1)
            self.certainties[(p,children[i])] = var

            if not children[i] in self.parents_of:
                self.parents_of[children[i]] = set()

            if not p in self.children_of:
                self.children_of[p] = set()

            self.parents_of[children[i]].add(p)
            self.children_of[p].add(children[i])

        #learn all additional children
        #learn first quartet
        #only need to consider children that are siblings with ALL of the initial quartet
        a,b,c,d = children
        candidate_children = set(self.siblings[a])
        for z in [b,c,d]:
            candidate_children &= self.siblings[z]

        candidate_children = sorted(set(range(self.numChildren)) - set((a,b,c,d)))
        while len(candidate_children) > 0:
            z = candidate_children.pop()
            if self.extendParent(p, z, (a,b,c,d)):
                self.children_of[p].add(z)
                self.parents_of[z].add(p)
            
        self.currentParent +=1
        return self.currentParent-1

    ###################################
    #MAIN STRUCTURE LEARNING FUNCTION #
    ###################################

    def parallel_learn_structure(self, thresholds, numProcs, max_rounds=100):
        start_time = time.time()
        converged = False
        round = -1

        
        (eigen_thresh,
        coupled_thresh,
        f_min, f_max, p_min, p_max) = thresholds

        self.f_max = f_max
        self.p_min = p_min
        self.p_max = p_max
        self.f_min = f_min
        self.coupled_thresh = coupled_thresh

        self.initialize_siblings()

        while not converged:
            round += 1
            if round > max_rounds:
                break

            self.current_round = round
            self.parents_in_round[round] = set()

            quartets = set()
            converged = True
        
            quartet_list = self.dist.listSets(4)
            D_list = itertools.imap(self.adjustedD, quartet_list)
            quartet_list = itertools.izip(D_list, quartet_list, [round]*len(quartet_list))
            quartet_list = list(quartet_list)
            print len(quartet_list), 'quartet_list'


            quartets = list()
            counter = 0

            thread_pool = Pool(numProcs)

            for res in thread_pool.imap_unordered(parallelEigenTest, quartet_list, chunksize=1000):
            #for arg in quartet_list:
                #res = parallelEigenTest(arg)
                (a,b,c,d), score, D = res
                #print a,b,c,d
                counter += 1
                if counter % 1000 == 0:
                    print counter

                quartets.append((score, (a,b,c,d), D))

            print 'there are', len(quartets), 'before filtering'
            quartets = sorted(quartets, key=lambda x: min(x[0][:4]))
            quartets = filter(lambda x: x[0][0] > coupled_thresh, quartets)
            quartets = filter(lambda x: x[0][1] > coupled_thresh, quartets)
            quartets = filter(lambda x: x[0][2] > coupled_thresh, quartets)
            quartets = filter(lambda x: x[0][3] > coupled_thresh, quartets)
            quartets = filter(lambda x: x[0][4] < eigen_thresh, quartets)

            quartets = sorted(quartets, key=lambda x: x[0][4])

            print "there are", len(quartets), "quartets that passed coupled filters"
            self.scores = {}

            for score, (a,b,c,d), D in quartets:
                print "considering parent", (a,b,c,d), "with score", score
                #print D
                p_id = self.learnNewParent((a,b,c,d), D, round)
                if p_id < 0:
                    print "rejected parent"
                    continue

                print "accepted parent", p_id
                print "children are", self.children_of[p_id]
                self.parents_in_round[round].add(p_id)
                self.singly_coupled[p_id] = (a,b,c,d)
                self.scores[p_id] = score
                converged = False
                #print score
                
            self.writeLayer(round)

        end_time = time.time()

        self.learn_time = end_time - start_time
        print "learn time", self.learn_time


    ###########
    # Display #
    ###########
    
    def writeLayer(self, round,outfile=None):
        print "layer writing"
        if outfile == None:
            outfile = file(self.results_dir+"/layer_"+str(round), 'w')

        if round == -1:
            rounds = range(len(self.parents_in_round))
        else:
            rounds = [round]

        for round in rounds:
            for i in self.parents_in_round[round]:
                #score = self.scores[i]
                prior = (self.edges[(i,None)],)
                quartet = self.singly_coupled[i]
                print >>outfile, prior[0],':',
                for j in self.children:
                    if j in self.children_of[i]:
                        print >>outfile, self.edges[(i,j)],
                    else:
                        print >>outfile, 1,
                print >>outfile, '#',
                for q in quartet:
                    print >>outfile, q,
                print >>outfile, ""
        return 0

